from torch import nn
from typing import Callable


def change_num_classes(model, num_classes):
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    return model

def neko_cfg(f_cfg: Callable):
    f_cfg._neko_cfg_ = True
    return f_cfg


from .dataset import BaseDataset
from .bucket import BaseBucket, FixedBucket, FixedCropBucket, RatioBucket, SizeBucket, RatioSizeBucket, LongEdgeBucket
from .utils import CycleData
from .label_loader import JsonLabelLoader, YamlLabelLoader, TXTLabelLoader, auto_label_loader
from .sampler import DistributedCycleSampler, get_sampler
from typing import Dict, Any


class DataGroup:
    def __init__(self, loader_dict:Dict[str, Any], loss_weights:Dict[str, float], cycle=True):
        self.loader_dict = loader_dict
        self.loss_weights = loss_weights
        self.cycle = cycle

    def __iter__(self):
        if self.cycle:
            self.data_iter_dict = {name:iter(CycleData(loader)) for name, loader in self.loader_dict.items()}
        else:
            self.data_iter_dict = {}
            for name, loader in self.loader_dict.items():
                loader.dataset.bucket.rest(0) # rest bucket
                self.data_iter_dict[name] = iter(loader)
        return self

    def __next__(self):
        if self.cycle:
            return {name:next(data_iter) for name, data_iter in self.data_iter_dict.items()}
        else:
            data_dict = {}
            for name, data_iter in self.data_iter_dict.items():
                try:
                    data_dict[name] = next(data_iter)
                except StopIteration:
                    pass
            if len(data_dict) == 0:
                raise StopIteration()
            return data_dict

    def __len__(self):
        return max([len(loader) for loader in self.loader_dict.values()])

    def get_dataset(self, name):
        return self.loader_dict[name].dataset

    def get_loss_weights(self, name):
        return self.loss_weights[name]
    
    def first_loader(self):
        return next(iter(self.loader_dict.values()))


import hashlib
import math
import os
import random
import re
from typing import Tuple, List, Any, Dict

import torch
from omegaconf import OmegaConf

size_mul = {'K': 1 << 10, 'M': 1 << 20, 'G': 1 << 30, 'T': 1 << 40}
size_key = 'TGMK'


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


def low_rank_approximate(weight, rank, clamp_quantile=0.99):
    if len(weight.shape) == 4:  # conv
        weight = weight.flatten(1)
        out_ch, in_ch, k1, k2 = weight.shape

    U, S, Vh = torch.linalg.svd(weight)
    U = U[:, :rank]
    S = S[:rank]
    U = U @ torch.diag(S)

    Vh = Vh[:rank, :]

    dist = torch.cat([U.flatten(), Vh.flatten()])
    hi_val = torch.quantile(dist, clamp_quantile)
    low_val = -hi_val

    U = U.clamp(low_val, hi_val)
    Vh = Vh.clamp(low_val, hi_val)

    if len(weight.shape) == 4:
        # U is (out_channels, rank) with 1x1 conv.
        U = U.reshape(U.shape[0], U.shape[1], 1, 1)
        # V is (rank, in_channels * kernel_size1 * kernel_size2)
        Vh = Vh.reshape(Vh.shape[0], in_ch, k1, k2)
    return U, Vh


def get_cfg_range(cfg_text: str):
    dy_cfg_f = 'ln'
    if cfg_text.find(':') != -1:
        cfg_text, dy_cfg_f = cfg_text.split(':')

    if cfg_text.find('-') != -1:
        l, h = cfg_text.split('-')
        return float(l), float(h), dy_cfg_f
    else:
        return float(cfg_text), float(cfg_text), dy_cfg_f


def to_validate_file(name):
    rstr = r"[\/\\\:\*\?\"\<\>\|]"  # '/ \ : * ? " < > |'
    new_title = re.sub(rstr, "_", name)  # 替换为下划线
    return new_title[:50]


def make_mask(start, end, length):
    mask = torch.zeros(length)
    mask[int(length * start):int(length * end)] = 1
    return mask.bool()


def get_file_name(file: str):
    return file.rsplit('.', 1)[0]


def get_file_ext(file: str):
    try:
        return file.rsplit('.', 1)[1].lower()
    except:
        return ''


def factorization(dimension: int, factor: int = -1) -> Tuple[int, int]:
    find_one = lambda x: len(x) - (x.rfind('1') + 1)
    dim_bin = bin(dimension)
    num = find_one(dim_bin)
    f_max = (len(dim_bin) - 3) >> 1 if factor < 0 else find_one(bin(factor))
    num = min(num, f_max)
    return dimension >> num, 1 << num


def isinstance_list(obj, cls_list):
    for cls in cls_list:
        if isinstance(obj, cls):
            return True
    return False


def net_path_join(*args):
    return '.'.join(args).strip('.').replace('..', '.')


def mgcd(*args):
    g = args[0]
    for s in args[1:]:
        g = math.gcd(g, s)
    return g


def size_to_int(size):
    return int(size[:-3]) * size_mul[size[-3]]


def int_to_size(size):
    for i, k in zip(range(40, 0, -10), size_key):
        if size >= 1 << i:
            return f'{size >> i}{k}iB'


def prepare_seed(seeds: List[int], device='cuda'):
    return [torch.Generator(device=device).manual_seed(s or random.randint(0, 1 << 30)) for s in seeds]


def hash_str(data):
    return hashlib.sha256(data.encode('utf-8')).hexdigest()


def format_number(num):
    if num >= 1e9:
        return f'{num / 1e9:.1f}B'
    elif num >= 1e6:
        return f'{num / 1e6:.1f}M'
    elif num >= 1e3:
        return f'{num / 1e3:.1f}K'
    else:
        return str(num)


def is_list(v):
    return OmegaConf.is_list(v) or isinstance(v, list)

def is_dict(v):
    return OmegaConf.is_dict(v) or isinstance(v, dict)

def addto_dictlist(dict_list: Dict[str, List], data: Dict[str, Any], v_proc=None):
    for k, v in data.items():
        if k not in dict_list:
            dict_list[k] = []
        if v_proc is not None:
            dict_list[k].append(v_proc(v))
        else:
            dict_list[k].append(v)
    return dict_list

def set_list_value(lst, index, value, default=None):
    if index >= len(lst):
        lst.extend([default] * (index - len(lst) + 1))
    lst[index] = value

def dict_parse_list(data: Dict[str, Any]):
    if isinstance(data, dict):
        if len(data)>0 and isinstance(next(iter(data)), int):
            return [dict_parse_list(data[i]) for i in range(len(data))]
        else:
            return {k: dict_parse_list(v) for k, v in data.items()}
    else:
        return data


import math
import os.path
import pickle
from concurrent.futures import ThreadPoolExecutor

import cv2
import numpy as np
from loguru import logger
from sklearn.cluster import KMeans
from tqdm import tqdm

from .base import BaseBucket
from ..utils import resize_crop_fix, pad_crop_fix
from ..handler import AutoSizeHandler


class RatioBucket(BaseBucket):
    can_shuffle = False
    handler = AutoSizeHandler()

    def __init__(self, target_area: int = 640*640, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None):
        self.target_area = target_area
        self.step_size = step_size
        self.num_bucket = num_bucket
        self.pre_build_bucket = pre_build_bucket

    def load_bucket(self, path):
        with open(path, 'rb') as f:
            data = pickle.load(f)
        self.buckets = data['buckets']
        self.size_buckets = data['size_buckets']
        self.idx_bucket_map = data['idx_bucket_map']
        self.data_len = data['data_len']

    def save_bucket(self, path):
        with open(path, 'wb') as f:
            pickle.dump({
                'buckets':self.buckets,
                'size_buckets':self.size_buckets,
                'idx_bucket_map':self.idx_bucket_map,
                'data_len':self.data_len,
            }, f)

    def build_buckets_from_ratios(self):
        logger.info('build buckets from ratios')
        size_low = int(math.sqrt(self.target_area/self.ratio_max))
        size_high = int(self.ratio_max*size_low)

        # SD需要边长是8的倍数
        size_low = (size_low//self.step_size)*self.step_size
        size_high = (size_high//self.step_size)*self.step_size

        data = []
        for w in range(size_low, size_high+1, self.step_size):
            for h in range(size_low, size_high+1, self.step_size):
                data.append([w*h, np.log2(w/h), w, h])  # 对比例取对数，更符合人感知，宽高相反的可以对称分布。
        data = np.array(data)

        error_area = np.abs(data[:, 0]-self.target_area)
        data_use = data[np.argsort(error_area)[:self.num_bucket*3], :]  # 取最小的num_bucket*3个

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514).fit(data_use[:, 1].reshape(-1, 1))
        labels = kmeans.labels_
        self.buckets = []  # [bucket_id:[file_idx,...]]
        ratios_log = []
        self.size_buckets = []
        for i in range(self.num_bucket):
            map_idx = np.where(labels == i)[0]
            m_idx = map_idx[np.argmin(np.abs(data_use[labels == i, 1]-np.median(data_use[labels == i, 1])))]
            # self.buckets[wh_hash(*data_use[m_idx, 2:])]=[]
            self.buckets.append([])
            ratios_log.append(data_use[m_idx, 1])
            self.size_buckets.append(data_use[m_idx, 2:].astype(int))
        ratios_log = np.array(ratios_log)
        self.size_buckets = np.array(self.size_buckets)

        # get images ratio
        def get_ratio(info):
            data, source = info
            w, h = source.get_image_size(data)
            ratio = np.log2(w/h)
            return ratio

        ratio_list = []
        self.source.return_source = True
        with ThreadPoolExecutor() as executor:
            for ratio in tqdm(executor.map(get_ratio, self.source), desc='get image info', total=len(self.source)):
                ratio_list.append(ratio)
        ratio_list = np.array(ratio_list)
        self.source.return_source = False

        # fill buckets with images w,h
        bucket_id = np.abs(ratio_list[:,None] - ratios_log[None,:]).argmin(axis=-1)
        self.idx_bucket_map = bucket_id
        for bidx in range(self.num_bucket):
            self.buckets.append(np.where(bucket_id == bidx)[0].tolist())
        logger.info('buckets info: '+', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))

    def build_buckets_from_images(self):
        logger.info('build buckets from images')

        def get_ratio(info):
            data, source = info
            w, h = source.get_image_size(data)
            ratio = np.log2(w/h)
            return ratio

        ratio_list = []
        self.source.return_source = True
        with ThreadPoolExecutor() as executor:
            for ratio in tqdm(executor.map(get_ratio, self.source), desc='get image info', total=len(self.source)):
                ratio_list.append(ratio)
        ratio_list = np.array(ratio_list)
        self.source.return_source = False

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514, verbose=True, tol=1e-3).fit(ratio_list.reshape(-1, 1))
        labels = kmeans.labels_
        ratios = 2**kmeans.cluster_centers_.reshape(-1)

        h_all = np.sqrt(self.target_area/ratios)
        w_all = h_all*ratios

        # SD需要边长是8的倍数
        h_all = (np.round(h_all/self.step_size)*self.step_size).astype(int)
        w_all = (np.round(w_all/self.step_size)*self.step_size).astype(int)
        self.size_buckets = list(zip(w_all, h_all))
        self.size_buckets = np.array(self.size_buckets)

        self.buckets = []  # [bucket_id:[file_idx,...]]
        self.idx_bucket_map = labels
        for bidx in range(self.num_bucket):
            self.buckets.append(np.where(labels == bidx)[0].tolist())
        logger.info('buckets info: '+', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))

    def build(self, bs: int, world_size: int, source: 'ComposeDataSource'):
        '''
        :param bs: batch_size * n_gpus * accumulation_step
        :param img_root_list:
        '''
        self.source = source
        self.bs = bs*world_size
        if self.pre_build_bucket and os.path.exists(self.pre_build_bucket):
            self.load_bucket(self.pre_build_bucket)
            return

        self._build()

        rs = np.random.RandomState(42)
        # make len(bucket)%bs==0
        self.data_len = 0
        for bidx, bucket in enumerate(self.buckets):
            rest = len(bucket)%self.bs
            if rest>0:
                bucket.extend(rs.choice(bucket, self.bs-rest))
            self.data_len += len(bucket)
            self.buckets[bidx] = np.array(bucket)

        if self.pre_build_bucket:
            self.save_bucket(self.pre_build_bucket)

    def rest(self, epoch):
        rs = np.random.RandomState(42+epoch)
        bucket_list = [x.copy() for x in self.buckets]
        # shuffle inter bucket
        for x in bucket_list:
            rs.shuffle(x)

        # shuffle of batches
        bucket_list = np.hstack(bucket_list).reshape(-1, self.bs).astype(int)
        rs.shuffle(bucket_list)

        self.idx_bucket = bucket_list.reshape(-1)

    def __getitem__(self, idx):
        file_idx = self.idx_bucket[idx]
        bucket_idx = self.idx_bucket_map[file_idx]
        datas = self.source[file_idx]
        datas['image_size'] = self.size_buckets[bucket_idx]
        return datas

    def __len__(self):
        return self.data_len

    @classmethod
    def from_ratios(cls, target_area: int = 640*640, step_size: int = 8, num_bucket: int = 10, ratio_max: float = 4,
                    pre_build_bucket: str = None, **kwargs):
        arb = cls(target_area, step_size, num_bucket, pre_build_bucket=pre_build_bucket)
        arb.ratio_max = ratio_max
        arb._build = arb.build_buckets_from_ratios
        return arb

    @classmethod
    def from_files(cls, target_area: int = 640*640, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None, **kwargs):
        arb = cls(target_area, step_size, num_bucket, pre_build_bucket=pre_build_bucket)
        arb._build = arb.build_buckets_from_images
        return arb

class SizeBucket(RatioBucket):
    handler = AutoSizeHandler(mode='pad')

    def __init__(self, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None):
        super().__init__(step_size=step_size, num_bucket=num_bucket, pre_build_bucket=pre_build_bucket)

    def build_buckets_from_images(self):
        '''
        根据图像尺寸聚类，不会resize图像，只有剪裁和填充操作。
        '''
        logger.info('build buckets from images size')

        def get_size(info):
            data, source = info
            w, h = source.get_image_size(data)
            return w, h

        size_list = []
        self.source.return_source = True
        with ThreadPoolExecutor() as executor:
            for w, h in tqdm(executor.map(get_size, self.source), desc='get image info', total=len(self.source)):
                size_list.append([w, h])
        size_list = np.array(size_list)
        self.source.return_source = False

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514).fit(size_list)
        labels = kmeans.labels_
        size_buckets = kmeans.cluster_centers_

        # SD需要边长是8的倍数
        self.size_buckets = (np.round(size_buckets/self.step_size)*self.step_size).astype(int)

        self.buckets = []  # [bucket_id:[file_idx,...]]
        self.idx_bucket_map = labels
        for bidx in range(self.num_bucket):
            self.buckets.append(np.where(labels==bidx)[0].tolist())
        logger.info('buckets info: '+', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))

    @classmethod
    def from_files(cls, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None, **kwargs):
        arb = cls(step_size, num_bucket, pre_build_bucket=pre_build_bucket)
        arb._build = arb.build_buckets_from_images
        return arb

class RatioSizeBucket(RatioBucket):
    def __init__(self, step_size: int = 8, num_bucket: int = 10, max_area:int=640*640, pre_build_bucket: str = None):
        super().__init__(step_size=step_size, num_bucket=num_bucket, pre_build_bucket=pre_build_bucket)
        self.max_area = max_area

    def build_buckets_from_images(self):
        '''
        根据图像尺寸聚类，不会resize图像，只有剪裁和填充操作。
        '''
        logger.info('build buckets from images')

        def get_ratio(info):
            data, source = info
            w, h = source.get_image_size(data)
            ratio = np.log2(w/h)
            log_area = np.log2(min(w * h, self.max_area))
            return ratio, log_area

        ratio_list = []
        self.source.return_source = True
        with ThreadPoolExecutor() as executor:
            for ratio, log_area in tqdm(executor.map(get_ratio, self.source), desc='get image info', total=len(self.source)):
                ratio_list.append([ratio, log_area])
        ratio_list = np.array(ratio_list)
        self.source.return_source = False

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514).fit(ratio_list)
        labels = kmeans.labels_
        ratios = 2**kmeans.cluster_centers_[:, 0]
        sizes = 2**kmeans.cluster_centers_[:, 1]

        h_all = np.sqrt(sizes/ratios)
        w_all = h_all*ratios

        # SD需要边长是8的倍数
        h_all = (np.round(h_all/self.step_size)*self.step_size).astype(int)
        w_all = (np.round(w_all/self.step_size)*self.step_size).astype(int)
        self.size_buckets = list(zip(w_all, h_all))
        self.size_buckets = np.array(self.size_buckets)

        self.buckets = []  # [bucket_id:[file_idx,...]]
        self.idx_bucket_map = labels
        for bidx in range(self.num_bucket):
            self.buckets.append(np.where(labels==bidx)[0].tolist())
        logger.info('buckets info: '+', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))

    @classmethod
    def from_files(cls, step_size: int = 8, num_bucket: int = 10, max_area:int=640*640, pre_build_bucket: str = None, **kwargs):
        arb = cls(step_size, num_bucket, max_area=max_area, pre_build_bucket=pre_build_bucket)
        arb._build = arb.build_buckets_from_images
        return arb


from typing import Tuple, Union

import cv2
import numpy as np

from .base import BaseBucket
from ..handler import AutoSizeHandler


class PosNegBucket(BaseBucket):
    can_shuffle = False
    handler = AutoSizeHandler(mode='full')

    def __init__(self, target_size: Union[Tuple[int, int], int] = 512, pos_rate=0.5, **kwargs):
        self.target_size = (target_size, target_size) if isinstance(target_size, int) else target_size
        self.pos_rate = pos_rate

    def build(self, bs: int, world_size: int, source: 'DataSource'):
        self.bs = bs  # bs per device
        self.world_size = world_size
        self.source = source

        self.cls_group = {}  # {cls: idx}
        for i, (path, source) in enumerate(self.source):
            cls_name = source.get_class_name(path)
            if cls_name not in self.cls_group:
                self.cls_group[cls_name] = []
            self.cls_group[cls_name].append(i)
        self.cls_names = list(self.cls_group.keys())

    def rest(self, epoch):
        self.rs = np.random.RandomState(42 + epoch)

        # shuffle of batches
        img_idxs = np.arange(0, len(self.source)).astype(int)
        self.rs.shuffle(img_idxs)

        self.img_idxs = img_idxs

    def __getitem__(self, idx) -> Tuple[Tuple[str, 'DataSource'], Tuple[int, int]]:

        # world_size for DistributedSampler
        ws_bs = self.world_size * self.bs
        idx_bs = (idx // self.world_size) % self.bs
        idx_0 = idx // ws_bs * ws_bs + idx % self.world_size

        if idx_bs == 0:
            return self.source[self.img_idxs[idx]], self.target_size
        elif idx_bs < int(self.bs * self.pos_rate):  # pos
            path, source = self.source[self.img_idxs[idx_0]]
            idx_c = self.rs.choice(self.cls_group[source.get_class_name(path)])
            return self.source[idx_c], self.target_size
        else:  # neg
            path, source = self.source[self.img_idxs[idx_0]]
            cls_name = source.get_class_name(path)
            cls_name_c = self.rs.choice(self.cls_names)
            while cls_name == cls_name_c:
                cls_name_c = self.rs.choice(self.cls_names)
            idx_c = self.rs.choice(self.cls_group[cls_name_c])
            return self.source[idx_c], self.target_size

    def __len__(self):
        return len(self.source)


class TripletBucket(PosNegBucket):
    def __getitem__(self, idx) -> Tuple[Tuple[str, 'DataSource'], Tuple[int, int]]:
        assert self.bs % 3 == 0, 'batch size of TripletBucket must be a multiple of 3.'
        bs_1 = self.bs // 3

        # world_size for DistributedSampler
        ws_bs = self.world_size * self.bs
        idx_bs = (idx // self.world_size) % self.bs
        idx_0 = idx // ws_bs * ws_bs + idx % self.world_size + (idx_bs % bs_1) * self.world_size
        
        if idx_bs < bs_1:
            return self.source[self.img_idxs[idx]], self.target_size
        elif idx_bs < bs_1 * 2:  # pos
            path, source = self.source[self.img_idxs[idx_0]]
            idx_c = self.rs.choice(self.cls_group[source.get_class_name(path)])
            return self.source[idx_c], self.target_size
        else:  # neg
            path, source = self.source[self.img_idxs[idx_0]]
            cls_name = source.get_class_name(path)
            cls_name_c = self.rs.choice(self.cls_names)
            while cls_name == cls_name_c:
                cls_name_c = self.rs.choice(self.cls_names)
            idx_c = self.rs.choice(self.cls_group[cls_name_c])
            return self.source[idx_c], self.target_size

    def process_label(self, idx:int, label):
        bs_1 = self.bs // 3
        ws_bs = self.world_size * self.bs
        idx_bs = (idx // self.world_size) % self.bs
        idx_0 = idx // ws_bs * ws_bs + idx % self.world_size + (idx_bs % bs_1) * self.world_size
        
        if idx_bs < bs_1:
            label['label']['label'] = 0
        elif idx_bs < bs_1 * 2:  # pos
            label['label']['label'] = 1
        else:  # neg
            label['label']['label'] = 2
            
        return label


import torch
from addict import Dict

from .workflow import MemoryMixin


class WorkflowRunner:
    def __init__(self):
        self.memory = Dict()

    @torch.inference_mode()
    def run(self, actions, states=None):
        if states is None:
            states = dict()
        N_steps = len(actions)
        for step, act in enumerate(actions):
            print(f'[{step + 1}/{N_steps}] action: {type(act).__name__}')
            if isinstance(act, MemoryMixin):
                states = act(memory=self.memory, **states)
            else:
                states = act(**states)
            print(f'states: {", ".join(states.keys())}')
        return states



from .ckpt_pkl import CkptManagerPKL
from .ckpt_safetensor import CkptManagerSafe

def auto_manager(ckpt_path:str, **kwargs):
    return CkptManagerSafe(**kwargs) if ckpt_path.endswith('.safetensors') else CkptManagerPKL(**kwargs)


from functools import partial

import torch
import torchvision
import torchvision.transforms as T
from torch import nn
from torch.nn import CrossEntropyLoss
from torchmetrics.classification import MulticlassAccuracy, MulticlassF1Score

from cfgs.py.train import train_base, tuning_base
from rainbowneko.ckpt_manager import CkptManagerPKL
from rainbowneko.evaluate import MetricGroup, MetricContainer, Evaluator
from rainbowneko.models.wrapper import SingleWrapper
from rainbowneko.parser import make_base, CfgModelParser
from rainbowneko.train.data import FixedBucket
from rainbowneko.train.data import BaseDataset
from rainbowneko.train.data.source import IndexSource
from rainbowneko.train.data.handler import HandlerChain, ImageHandler, LoadImageHandler
from rainbowneko.train.loss import LossContainer
from rainbowneko.utils import neko_cfg

num_classes = 10

def load_resnet():
    model = torchvision.models.resnet18(weights=torchvision.models.ResNet18_Weights.DEFAULT)
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    return model

def make_cfg():
    dict(
        _base_=make_base(train_base, tuning_base)+[],

        model_part=CfgModelParser([
            dict(
                lr=1e-2,
                layers=[''],  # train all layers
            )
        ]),

        # func(_partial_=True, ...) same as partial(func, ...)
        ckpt_manager=CkptManagerPKL(_partial_=True, saved_model=(
            {'model':'model', 'trainable':False},
        )),

        train=dict(
            train_epochs=100,
            workers=2,
            max_grad_norm=None,
            save_step=2000,

            loss=LossContainer(loss=CrossEntropyLoss()),

            optimizer=partial(torch.optim.AdamW, weight_decay=5e-4),

            scale_lr=False,
            scheduler=dict(
                name='cosine',
                num_warmup_steps=10,
            ),
            metrics=MetricGroup(metric_dict=dict(
                acc=MetricContainer(MulticlassAccuracy(num_classes=num_classes)),
                f1=MetricContainer(MulticlassF1Score(num_classes=num_classes)),
            )),
        ),

        model=dict(
            name='cifar-resnet18',
            wrapper=partial(SingleWrapper, model=load_resnet())
        ),

        data_train=cfg_data(), # config can be split into another function with @neko_cfg

        evaluator=cfg_evaluator(),
    )

@neko_cfg
def cfg_data():
    dict(
        dataset1=partial(BaseDataset, batch_size=128, loss_weight=1.0,
            source=dict(
                data_source1=IndexSource(
                    data=torchvision.datasets.cifar.CIFAR10(root=r'D:\others\dataset\cifar', train=True, download=True)
                ),
            ),
            handler=HandlerChain(handlers=dict(
                load=LoadImageHandler(),
                bucket=FixedBucket.handler, # bucket 会自带一些处理模块
                image=ImageHandler(transform=T.Compose([
                        T.RandomCrop(size=32, padding=4),
                        T.RandomHorizontalFlip(),
                        T.ToTensor(),
                        T.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010]),
                    ]),
                )
            )),
            bucket=FixedBucket(target_size=32),
        )
    )

@neko_cfg
def cfg_evaluator():
    partial(Evaluator,
        interval=500,
        metric=MetricGroup(metric_dict=dict(
            acc=MetricContainer(MulticlassAccuracy(num_classes=num_classes)),
            f1=MetricContainer(MulticlassF1Score(num_classes=num_classes)),
        )),
        dataset=dict(
            dataset1=partial(BaseDataset, batch_size=128, loss_weight=1.0,
                source=dict(
                    data_source1=IndexSource(
                        data=torchvision.datasets.cifar.CIFAR10(root=r'D:\others\dataset\cifar', train=False, download=True)
                    ),
                ),
                handler=HandlerChain(handlers=dict(
                    load=LoadImageHandler(),
                    bucket=FixedBucket.handler,
                    image=ImageHandler(transform=T.Compose([
                            T.ToTensor(),
                            T.Normalize(mean=[0.4914, 0.4822, 0.4465], std=[0.2023, 0.1994, 0.2010]),
                        ]),
                    )
                )),
                bucket=FixedBucket(target_size=32),
            )
        )
    )


from typing import Dict, Union

import torch
from tqdm.auto import tqdm

from rainbowneko.models.wrapper import BaseWrapper
from rainbowneko.train.data import DataGroup
from rainbowneko.utils import is_dict
from .metrics import BaseMetric, MetricGroup


class Evaluator:
    def __init__(self, trainer: "Trainer", data_loader_group: DataGroup, metric: Union[BaseMetric, Dict[str, BaseMetric]], interval=100):
        self.data_loader_group = data_loader_group
        self.trainer = trainer
        self.metric = metric
        self.interval = interval

        if is_dict(self.metric):
            for metric in self.metric.values():
                if metric is not None:
                    metric.to(self.device)
        else:
            self.metric.to(self.device)

    @property
    def device(self):
        return self.trainer.device

    def forward_one_step(self, ds_name, model, data):
        input_datas = {k: self.trainer.to_dev(v) for k, v in data.items() if k != "plugin_input"}
        if "plugin_input" in data:
            input_datas["plugin_input"] = {k: self.trainer.to_dev(v) for k, v in data["plugin_input"].items()}

        model_pred = model(ds_name, **input_datas)

        return model_pred, input_datas

    @torch.inference_mode()
    def evaluate(self, step:int, model: BaseWrapper):
        if step % self.interval != 0:
            return

        # reset metric
        model.eval()
        if is_dict(self.metric):
            for metric in self.metric.values():
                if metric is not None:
                    metric.reset()
        else:
            self.metric.reset()

        for data_dict in tqdm(self.data_loader_group, disable=not self.trainer.is_local_main_process):
            for ds_name, data in data_dict.items():
                pred, input_datas = self.forward_one_step(ds_name, model, data)
                # update data to metric
                if is_dict(self.metric):
                    if self.metric[ds_name] is not None:
                        self.metric[ds_name].update(pred, input_datas)
                else:
                    self.metric.update(pred, input_datas)

        if is_dict(self.metric):
            v_metric = {}
            for ds_name, metric in self.metric.items():
                if metric is not None:
                    v_metric_i = metric.finish(self.trainer.accelerator.gather, self.trainer.is_local_main_process)
                    if not isinstance(v_metric_i, dict):
                        v_metric_i = {'metric': v_metric_i}
                    v_metric.update({f'{ds_name}/{k}': v for k, v in v_metric_i.items()})
        else:
            v_metric = self.metric.finish(self.trainer.accelerator.gather, self.trainer.is_local_main_process)
            if not isinstance(v_metric, dict):
                v_metric = {'metric': v_metric}

        data_size = {name:len(loader.dataset) for name, loader in self.data_loader_group.loader_dict.items()}
        self.trainer.loggers.info(f'Evaluate: data size {data_size}')
        log_data = {
            "eval/Step": {
                "format": "{}",
                "data": [self.trainer.global_step],
            }
        }
        log_data.update(MetricGroup.format(v_metric, prefix='eval/'))
        self.trainer.loggers.log(log_data, self.trainer.global_step)

    def to(self, device):
        self.metric.to(device)

class EvaluatorGroup:
    def __init__(self, loggers, evaluator_dict:Dict[str, Evaluator]):
        self.loggers = loggers
        self.evaluator_dict = evaluator_dict

    def evaluate(self, step:int, model: BaseWrapper):
        for name, evaluator in self.evaluator_dict.items():
            self.loggers.info(f'Evaluator {name}:')
            evaluator.evaluate(step, model)

    def to(self, device):
        for evaluator in self.evaluator_dict.values():
            evaluator.to(device)



from .base import DataSource, ComposeDataSource
from .index import IndexSource
from .img_label import ImageLabelSource
from .img_pair import ImagePairSource
from .folder_class import ImageFolderClassSource
from .unlabel import UnLabelSource


from typing import Tuple, Any, Dict

import cv2


class BaseBucket:
    can_shuffle = True

    def __getitem__(self, idx) -> Dict[str, Any]:
        '''
        :return: (file name of image), (target image size)
        '''
        return self.source[idx]

    def __len__(self):
        return len(self.source)

    def build(self, bs: int, world_size: int, source: 'DataSource'):
        self.source = source

    def rest(self, epoch):
        pass

    def process_label(self, idx:int, label):
        return label


"""
cfg_net_tools.py
====================
    :Name:        creat model and plugin from config
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     10/03/2023
    :Licence:     Apache-2.0
"""

import re
from typing import Dict, List, Tuple, Union, Any

import torch
from torch import nn
import functools

from rainbowneko.models.plugin import SinglePluginBlock, MultiPluginBlock, PluginBlock, PluginGroup, PatchPluginBlock
from rainbowneko.ckpt_manager import CkptManagerPKL, CkptManagerSafe
from ..utils import net_path_join
from ..utils import split_module_name


def get_class_match_layer(class_name, block: nn.Module):
    if type(block).__name__ == class_name:
        return [""]
    else:
        return ["." + name for name, layer in block.named_modules() if type(layer).__name__ == class_name]


def get_match_layers(layers, all_layers, return_metas=False) -> Union[List[str], List[Dict[str, Any]]]:
    res = []
    for name in layers:
        metas = name.split(":")

        use_re = False
        pre_hook = False
        cls_filter = None
        for meta in metas[:-1]:
            if meta == "re":
                use_re = True
            elif meta == "pre_hook":
                pre_hook = True
            elif meta.startswith("cls("):
                cls_filter = meta[4:-1]

        name = metas[-1]
        if use_re:
            pattern = re.compile(name)
            match_layers = filter(lambda x: pattern.match(x) != None, all_layers.keys())
        else:
            match_layers = [name]

        if cls_filter is not None:
            match_layers_new = []
            for layer in match_layers:
                match_layers_new.extend([layer + x for x in get_class_match_layer(name[1], all_layers[layer])])
            match_layers = match_layers_new

        for layer in match_layers:
            if return_metas:
                res.append({"layer": layer, "pre_hook": pre_hook})
            else:
                res.append(layer)

    # Remove duplicates and keep the original order
    if return_metas:
        layer_set = set()
        res_unique = []
        for item in res:
            if item["layer"] not in layer_set:
                layer_set.add(item["layer"])
                res_unique.append(item)
        return res_unique
    else:
        return sorted(set(res), key=res.index)


class CfgModelParser:
    def __init__(self, cfg_model, lr=1e-5, weight_decay=0):
        self.cfg_model = cfg_model
        self.lr = lr
        self.weight_decay = weight_decay

    def get_params_group(self, model):
        named_modules = {k: v for k, v in model.named_modules()}

        train_params = []
        train_layers = []

        if self.cfg_model is not None:
            for item in self.cfg_model:
                params_group = []
                for layer_name in get_match_layers(item.layers, named_modules):
                    layer = named_modules[layer_name]
                    layer.requires_grad_(True)
                    layer.train()
                    train_layers.append(layer)
                    params_group.extend(layer.parameters())
                train_params.append({"params": list(dict.fromkeys(params_group)), "lr": getattr(item, "lr", self.lr),
                                    "weight_decay": getattr(item, "weight_decay", self.weight_decay)})

        return train_params, train_layers

class CfgWDModelParser(CfgModelParser):
    def get_params_group(self, model):
        named_modules = {k: v for k, v in model.named_modules()}

        train_params = []
        train_layers = []

        wd_params = set()
        for m in model.modules():
            if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
                wd_params.add(m.weight)

        if self.cfg_model is not None:
            for item in self.cfg_model:
                params_group = []
                params_group_wd = []
                for layer_name in get_match_layers(item.layers, named_modules):
                    layer = named_modules[layer_name]
                    layer.requires_grad_(True)
                    layer.train()
                    train_layers.append(layer)
                    for p in layer.parameters():
                        if p in wd_params:
                            params_group_wd.append(p)
                        else:
                            params_group.append(p)
                if len(params_group)>0:
                    train_params.append({"params": list(dict.fromkeys(params_group)), "lr": getattr(item, "lr", self.lr),
                                        "weight_decay": 0})
                if len(params_group_wd) > 0:
                    train_params.append({"params": list(dict.fromkeys(params_group_wd)), "lr": getattr(item, "lr", self.lr),
                                         "weight_decay": getattr(item, "weight_decay", self.weight_decay)})


        return train_params, train_layers


class CustomModelParser:
    def __init__(self, params_loader, **kwargs):
        self.params_loader = params_loader
        self.kwargs = kwargs

    def get_params_group(self, model):
        train_params, train_layers = self.params_loader(model, **self.kwargs)
        return train_params, train_layers

def parse_plugin_cfg(model, cfg_plugin, default_lr=1e-5) -> Tuple[List, Dict[str, PluginGroup]]:
    train_params = []
    all_plugin_group = {}

    if cfg_plugin is None:
        return train_params, all_plugin_group

    named_modules = {k: v for k, v in model.named_modules()}

    for plugin_name, builder in cfg_plugin.items():
        builder: functools.partial
        all_plugin_blocks = {}

        lr = builder.keywords.pop("lr") if "lr" in builder.keywords else default_lr
        train_plugin = builder.keywords.pop("train") if "train" in builder.keywords else True
        plugin_class = getattr(builder.func, "__self__", builder.func)  # support static or class method

        params_group = []
        if issubclass(plugin_class, MultiPluginBlock):
            from_layers = [
                {**item, "layer": named_modules[item["layer"]]}
                for item in get_match_layers(builder.keywords.pop("from_layers"), named_modules, return_metas=True)
            ]
            to_layers = [
                {**item, "layer": named_modules[item["layer"]]}
                for item in get_match_layers(builder.keywords.pop("to_layers"), named_modules, return_metas=True)
            ]

            layer = builder(name=plugin_name, host_model=model, from_layers=from_layers, to_layers=to_layers)
            if train_plugin:
                layer.train()
                params = layer.get_trainable_parameters()
                for p in params:
                    p.requires_grad_(True)
                    params_group.append(p)
            else:
                layer.requires_grad_(False)
                layer.eval()
            all_plugin_blocks[""] = layer
        elif issubclass(plugin_class, SinglePluginBlock):
            layers_name = builder.keywords.pop("layers")
            for layer_name in get_match_layers(layers_name, named_modules):
                blocks = builder(name=plugin_name, host_model=model, host=named_modules[layer_name])
                if not isinstance(blocks, dict):
                    blocks = {"": blocks}

                for k, v in blocks.items():
                    all_plugin_blocks[net_path_join(layer_name, k)] = v
                    if train_plugin:
                        v.train()
                        params = v.get_trainable_parameters()
                        for p in params:
                            p.requires_grad_(True)
                            params_group.append(p)
                    else:
                        v.requires_grad_(False)
                        v.eval()
        elif issubclass(plugin_class, PluginBlock):
            from_layer = get_match_layers(builder.keywords.pop("from_layer"), named_modules, return_metas=True)
            to_layer = get_match_layers(builder.keywords.pop("to_layer"), named_modules, return_metas=True)

            for from_layer_meta, to_layer_meta in zip(from_layer, to_layer):
                from_layer_name = from_layer_meta["layer"]
                from_layer_meta["layer"] = named_modules[from_layer_name]
                to_layer_meta["layer"] = named_modules[to_layer_meta["layer"]]
                layer = builder(name=plugin_name, host_model=model, from_layer=from_layer_meta, to_layer=to_layer_meta)
                if train_plugin:
                    layer.train()
                    params = layer.get_trainable_parameters()
                    for p in params:
                        p.requires_grad_(True)
                        params_group.append(p)
                else:
                    layer.requires_grad_(False)
                    layer.eval()
                all_plugin_blocks[from_layer_name] = layer
        elif issubclass(plugin_class, PatchPluginBlock):
            layers_name = builder.keywords.pop("layers")
            for layer_name in get_match_layers(layers_name, named_modules):
                parent_name, host_name = split_module_name(layer_name)
                layers = builder(
                    name=plugin_name,
                    host_model=model,
                    host=named_modules[layer_name],
                    parent_block=named_modules[parent_name],
                    host_name=host_name,
                )
                if not isinstance(layers, dict):
                    layers = {"": layers}

                for k, v in layers.items():
                    all_plugin_blocks[net_path_join(layer_name, k)] = v
                    if train_plugin:
                        v.train()
                        params = v.get_trainable_parameters()
                        for p in params:
                            p.requires_grad_(True)
                            params_group.append(p)
                    else:
                        v.requires_grad_(False)
                        v.eval()
        else:
            raise NotImplementedError(f"Unknown plugin {plugin_class}")
        if train_plugin:
            train_params.append({"params": params_group, "lr": lr})
        all_plugin_group[plugin_name] = PluginGroup(all_plugin_blocks)
    return train_params, all_plugin_group


@torch.no_grad()
def load_models(model: nn.Module, cfg_merge):
    named_modules = {k: v for k, v in model.named_modules()}
    named_params = {k: v for k, v in model.named_parameters()}
    all_lora_blocks = {}

    ckpt_manager_torch = CkptManagerPKL()
    ckpt_manager_safe = CkptManagerSafe()

    def get_ckpt_manager(path: str):
        return ckpt_manager_safe if path.endswith(".safetensors") else ckpt_manager_torch

    if getattr(cfg_merge, "lora", None) is not None:
        for lora_id, item in enumerate(cfg_merge.lora):
            lora_state = get_ckpt_manager(item.path).load_ckpt(item.path, map_location="cpu")["lora"]
            lora_block_state = {}
            # get all layers in the lora_state
            for name, p in lora_state.items():
                # lora_block. is the old format
                prefix, block_name = name.split(".___." if name.rfind("lora_block.") == -1 else ".lora_block.", 1)
                if prefix not in lora_block_state:
                    lora_block_state[prefix] = {}
                lora_block_state[prefix][block_name] = p
            # get selected layers
            if item.layers != "all":
                match_blocks = get_match_layers(item.layers, named_modules)
                lora_state_new = {}
                for k, v in lora_block_state.items():
                    for mk in match_blocks:
                        if k.startswith(mk):
                            lora_state_new[k] = v
                            break
                lora_block_state = lora_state_new
            # add lora to host and load weights
            for host_name, lora_state in lora_block_state.items():
                lora_layer_cls, rank, rank_groups = get_lora_rank_and_cls(lora_state)
                if "alpha" in lora_state:
                    del lora_state["alpha"]
                if "scale" in lora_state:  # old format
                    del lora_state["scale"]

                lora_block = lora_layer_cls.wrap_layer(
                    lora_id,
                    named_modules[host_name],
                    rank=rank,
                    dropout=getattr(item, "dropout", 0.0),
                    alpha=getattr(item, "alpha", 1.0),
                    bias="layer.lora_up.bias" in lora_state,
                    rank_groups=rank_groups,
                    alpha_auto_scale=getattr(item, "alpha_auto_scale", True),
                )
                all_lora_blocks[f"{host_name}.{lora_block.name}"] = lora_block
                lora_block.load_state_dict(lora_state, strict=False)
                lora_block.set_mask(getattr(item, "mask", None))
                lora_block.to(model.device)

    if getattr(cfg_merge, "part", None) is not None:
        for item in cfg_merge.part:
            part_state = get_ckpt_manager(item.path).load_ckpt(item.path, map_location="cpu")["base"]
            if item.layers == "all":
                for k, v in part_state.items():
                    named_params[k].data = cfg_merge.base_model_alpha * named_params[k].data + item.alpha * v
            else:
                match_blocks = get_match_layers(item.layers, named_modules)
                state_add = {k: v for blk in match_blocks for k, v in part_state.items() if k.startswith(blk)}
                for k, v in state_add.items():
                    named_params[k].data = cfg_merge.base_model_alpha * named_params[k].data + item.alpha * v

    if getattr(cfg_merge, "plugin", None) is not None:
        for name, item in cfg_merge.plugin.items():
            plugin_state = get_ckpt_manager(item.path).load_ckpt(item.path, map_location="cpu")["plugin"]
            if item.layers != "all":
                match_blocks = get_match_layers(item.layers, named_modules)
                plugin_state = {k: v for blk in match_blocks for k, v in plugin_state.items() if k.startswith(blk)}
            plugin_key_set = set([k.split("___", 1)[0] + name for k in plugin_state.keys()])
            plugin_state = {
                k.replace("___", name): v for k, v in plugin_state.items()
            }  # replace placeholder to target plugin name
            model.load_state_dict(plugin_state, strict=False)
            del item.layers
            del item.path
            if hasattr(model, name):  # MultiPluginBlock
                getattr(model, name).set_hyper_params(**item)
            else:
                for plugin_key in plugin_key_set:
                    named_modules[plugin_key].set_hyper_params(**item)

    return LoraGroup(all_lora_blocks)



from .cfg_resolvers import *
from .cfg_tool import change_num_classes, neko_cfg
from .module import disable_hf_loggers
from .net import (get_scheduler, get_scheduler_with_name, remove_layers, remove_all_hooks, to_cpu, to_cuda, hook_compile,
                  split_module_name)
from .utils import *
from .key_mapper import KeyMapper
from .random import RandomContext


from typing import Union, Dict, Any


class CkptManagerBase:
    def __init__(self, **kwargs):
        pass

    def set_save_dir(self, save_dir):
        raise NotImplementedError()

    def save(self, name, step, model, all_plugin, ema={}, **kwargs):
        raise NotImplementedError()

    @classmethod
    def load(cls, pretrained_model, **kwargs):
        raise NotImplementedError

    @classmethod
    def load_to_model(cls, model, ckpt_path: Union[str, Dict[str, Any]], **kwargs):
        raise NotImplementedError



from typing import Dict, Any, Callable

import torch

from .base import BasicAction, feedback_input, MemoryMixin


class PrepareAction(BasicAction):
    def __init__(self, device='cuda', dtype=torch.float32):
        super().__init__()
        self.device = device
        self.dtype = dtype

    @feedback_input
    def forward(self, **states):
        return {'device': self.device, 'dtype': self.dtype}


class BuildModelAction(BasicAction, MemoryMixin):
    def __init__(self, model_builder: Callable):
        super().__init__()
        self.model_builder = model_builder

    @feedback_input
    def forward(self, memory, device, **states):
        memory.model = self.model_builder().to(device)


class ForwardAction(BasicAction, MemoryMixin):

    @feedback_input
    def forward(self, input: Dict[str, Any], memory, **states):
        with torch.inference_mode():
            output: Dict[str, Any] = memory.model(**input)
        return {'output': output}


class VisPredAction(BasicAction):
    @feedback_input
    def forward(self, output: Dict[str, Any], **states):
        pred = output['pred']
        if isinstance(pred, torch.Tensor):
            print(pred.cpu())
        else:
            print(pred)



import os
from .cfg_net_parser import parse_plugin_cfg, CfgModelParser, CustomModelParser, CfgWDModelParser
from .yaml_cfg import YamlCfgParser
from .python_cfg import PythonCfgParser

def load_config(path: str, remove_undefined=True):
    if path.lower().endswith('.yaml'):
        parser = YamlCfgParser()
    elif path.lower().endswith('.py'):
        parser = PythonCfgParser()
    else:
        raise ValueError('Unsupported config file format: {}'.format(path))
    return parser, parser.load_config(path, remove_undefined)

def load_config_with_cli(path: str, args_list=None, remove_undefined=True):
    if path.lower().endswith('.yaml'):
        parser = YamlCfgParser()
    elif path.lower().endswith('.py'):
        parser = PythonCfgParser()
    else:
        raise ValueError('Unsupported config file format: {}'.format(path))
    return parser, parser.load_config_with_cli(path, args_list, remove_undefined)

def get_rel_path(path):
    if not isinstance(path, str):
        path = path.__file__
    current_dir = os.getcwd()
    return os.path.relpath(path, current_dir)

def make_base(*base):
    return [get_rel_path(x) for x in base]


"""
train_ac.py
====================
    :Name:        train with accelerate
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     10/03/2023
    :Licence:     Apache-2.0
"""

import argparse
import math
import os
import warnings
from functools import partial

import hydra
import torch
# fix checkpoint bug for train part of model
import torch.utils.checkpoint
import torch.utils.data
from accelerate import Accelerator
from accelerate.utils import set_seed
from rainbowneko.evaluate import EvaluatorGroup, MetricGroup
from rainbowneko.models.wrapper import BaseWrapper
from rainbowneko.parser import load_config_with_cli
from rainbowneko.parser import parse_plugin_cfg
from rainbowneko.train.data import RatioBucket, DataGroup, get_sampler
from rainbowneko.train.loggers import LoggerGroup
from rainbowneko.utils import get_scheduler, mgcd, format_number, disable_hf_loggers, addto_dictlist, is_dict
from tqdm import tqdm

try:
    import xformers

    xformers_available = True
except:
    xformers_available = False


class Trainer:
    weight_dtype_map = {"fp32": torch.float32, "fp16": torch.float16, "bf16": torch.bfloat16}

    def __init__(self, parser, cfgs_raw):
        torch.backends.cudnn.benchmark = True

        cfgs = hydra.utils.instantiate(cfgs_raw)
        self.cfgs = cfgs
        self.parser = parser

        self.init_context(cfgs_raw)
        self.build_loggers(cfgs_raw)

        self.build_ckpt_manager()
        self.build_model()
        self.make_hooks()
        self.config_model()

        # build dataset
        self.batch_size_list = []
        assert len(cfgs.data_train) > 0, "At least one dataset is need."
        loss_weights = {name:dataset.keywords["loss_weight"] for name, dataset in cfgs.data_train.items()}
        self.train_loader_group = DataGroup(
            {name:self.build_data(dataset, train=True) for name, dataset in cfgs.data_train.items()}, loss_weights
        )

        # calculate steps and epochs
        self.steps_per_epoch = len(self.train_loader_group.first_loader())
        if self.cfgs.train.train_epochs is not None:
            self.cfgs.train.train_steps = self.cfgs.train.train_epochs * self.steps_per_epoch
        else:
            self.cfgs.train.train_epochs = math.ceil(self.cfgs.train.train_steps / self.steps_per_epoch)

        self.build_optimizer_scheduler()
        self.build_loss()

        with torch.no_grad():
            self.build_ema()

        self.load_resume()

        torch.backends.cuda.matmul.allow_tf32 = cfgs.allow_tf32

        if self.cfgs.train.metrics is not None:
            self.metric_train = self.cfgs.train.metrics
            if is_dict(self.metric_train):
                for metric in self.metric_train.values():
                    if metric is not None:
                        metric.to(self.accelerator.device)
            else:
                self.metric_train.to(self.accelerator.device)
        else:
            self.metric_train = None

        if self.cfgs.evaluator is not None:
            self.evaluator = self.build_evaluator(self.cfgs.evaluator)
            self.evaluator.to(self.accelerator.device)
        else:
            self.evaluator = None

        self.prepare()

    @property
    def device(self):
        return self.accelerator.device

    @property
    def is_local_main_process(self):
        return self.accelerator.is_local_main_process

    def init_context(self, cfgs_raw):
        self.accelerator = Accelerator(
            gradient_accumulation_steps=self.cfgs.train.gradient_accumulation_steps,
            mixed_precision=self.cfgs.mixed_precision,
            step_scheduler_with_optimizer=False,
        )

        self.local_rank = int(os.environ.get("LOCAL_RANK", -1))
        self.world_size = self.accelerator.num_processes

        set_seed(self.cfgs.seed + self.local_rank)

    def build_loggers(self, cfgs_raw):
        if self.is_local_main_process:
            self.exp_dir = self.cfgs.exp_dir
            os.makedirs(os.path.join(self.exp_dir, "ckpts/"), exist_ok=True)
            self.parser.save_configs(cfgs_raw, self.exp_dir)
            self.loggers: LoggerGroup = LoggerGroup([builder(exp_dir=self.exp_dir) for builder in self.cfgs.logger])
        else:
            self.loggers: LoggerGroup = LoggerGroup([builder(exp_dir=None) for builder in self.cfgs.logger])

        self.min_log_step = mgcd(*([item.log_step for item in self.loggers.logger_list]))

        self.loggers.info(f"world size (num GPUs): {self.world_size}")
        self.loggers.info(f"accumulation: {self.cfgs.train.gradient_accumulation_steps}")

        disable_hf_loggers(self.is_local_main_process)

    def prepare(self):
        # Prepare everything with accelerator.
        prepare_name_list, prepare_obj_list = [], []
        for k, v in self.model_wrapper.trainable_models.items():
            prepare_obj_list.append(v)
            prepare_name_list.append(k)
        N_trainable_models = len(prepare_obj_list)

        if hasattr(self, "optimizer"):
            prepare_obj_list.extend([self.optimizer, self.lr_scheduler] if self.lr_scheduler else [self.optimizer])
            prepare_name_list.extend(["optimizer", "lr_scheduler"] if self.lr_scheduler else ["optimizer"])

        prepare_obj_list.extend(self.train_loader_group.loader_dict.values())
        prepared_obj = self.accelerator.prepare(*prepare_obj_list)

        # prepared model
        if prepare_name_list[0] == "self":  # entire model is trainable
            self.model_wrapper = prepared_obj[0]
        else:
            for name, obj in zip(prepare_name_list[:N_trainable_models], prepared_obj[:N_trainable_models]):
                setattr(self.model_wrapper, name, obj)
        prepare_name_list = prepare_name_list[N_trainable_models:]
        prepared_obj = prepared_obj[N_trainable_models:]

        # prepared dataset
        ds_num = len(self.train_loader_group.loader_dict)
        for i, (k,v) in enumerate(self.train_loader_group.loader_dict.items()):
            self.train_loader_group.loader_dict[k] = prepared_obj[-ds_num+i]
        prepared_obj = prepared_obj[:-ds_num]

        # prepared optimizer and scheduler
        for name, obj in zip(prepare_name_list, prepared_obj):
            setattr(self, name, obj)

        self.model_wrapper: BaseWrapper = self.model_wrapper.to(self.device)
        if self.cfgs.model.force_cast_precision:
            self.model_wrapper.to(dtype=self.weight_dtype)

    def scale_lr(self, parameters):
        bs = sum(self.batch_size_list)
        scale_factor = bs * self.world_size * self.cfgs.train.gradient_accumulation_steps
        for param in parameters:
            if "lr" in param:
                param["lr"] *= scale_factor

    def build_model(self):
        self.model_wrapper = self.cfgs.model.wrapper()

    def build_ema(self):
        if self.cfgs.model.ema is not None:
            self.ema_model = self.cfgs.model.ema(self.model_wrapper.named_parameters())

    def build_loss(self):
        criterion = self.cfgs.train.loss
        if is_dict(criterion):
            self.criterion = {name:loss.to(self.device) for name, loss in criterion.items()}
        else:
            self.criterion = criterion.to(self.device)

    def build_ckpt_manager(self):
        self.ckpt_manager = self.cfgs.ckpt_manager()
        if self.is_local_main_process:
            self.ckpt_manager.set_save_dir(
                os.path.join(self.exp_dir, "ckpts"),
            )

    def build_evaluator(self, cfgs_eval):
        def build_one(cfgs_eval_one):
            dataset_cfg = cfgs_eval_one.keywords.pop('dataset', None)
            val_loader_group = DataGroup(
                {name:self.build_data(dataset, train=False) for name, dataset in dataset_cfg.items()}, None,
                cycle=False
            )

            evaluator = cfgs_eval_one(trainer=self, data_loader_group=val_loader_group)
            return evaluator

        if cfgs_eval is not None:
            if is_dict(cfgs_eval):
                evaluator_dict = {name:build_one(cfg) for name, cfg in cfgs_eval.items()}
                return EvaluatorGroup(loggers=self.loggers, evaluator_dict=evaluator_dict)
            else:
                return build_one(cfgs_eval)
        else:
            return None

    @property
    def model_raw(self):
        return self.model_wrapper.module

    def config_model(self):
        if self.cfgs.model.enable_xformers:
            if xformers_available:
                self.model_wrapper.enable_xformers()
            else:
                warnings.warn("xformers is not available. Make sure it is installed correctly")

        self.model_wrapper.requires_grad_(False)
        self.model_wrapper.eval()
        self.weight_dtype = self.weight_dtype_map.get(self.cfgs.mixed_precision, torch.float32)

        if self.cfgs.model.gradient_checkpointing:
            self.model_wrapper.enable_gradient_checkpointing()

    @torch.no_grad()
    def load_resume(self):
        if self.cfgs.train.resume is not None:
            for ckpt in self.cfgs.train.resume.ckpt_path:
                self.ckpt_manager.load_ckpt_to_model(
                    self.model_wrapper, ckpt, model_ema=getattr(self, "ema_model", None)
                )

    def make_hooks(self):
        pass

    def to_dev(self, x):
        if isinstance(x, torch.Tensor):
            if torch.is_floating_point(x):
                return x.to(self.device, dtype=self.weight_dtype)
            else:
                return x.to(self.device)
        else:
            return x

    def build_dataset(self, data_builder: partial):
        batch_size = data_builder.keywords.pop("batch_size")
        self.batch_size_list.append(batch_size)

        dataset = data_builder()
        dataset.build_bucket(bs=batch_size, world_size=self.world_size)
        self.loggers.info(f"len(dataset): {len(dataset)}")

        return dataset, batch_size

    def build_data(self, data_builder: partial, train=True) -> torch.utils.data.DataLoader:
        drop_last = data_builder.keywords.pop("drop_last", True)
        dataset, batch_size = self.build_dataset(data_builder)

        # Pytorch Data loader
        sampler = get_sampler(train)(
            dataset,
            num_replicas=self.world_size,
            rank=self.local_rank,
            shuffle=train and dataset.bucket.can_shuffle,
        )
        loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=self.cfgs.train.workers,
            sampler=sampler,
            collate_fn=dataset.collate_fn,
            drop_last=drop_last,
        )
        return loader

    def get_param_group_train(self):
        # make model part and plugin
        train_params, train_layers = self.cfgs.model_part.get_params_group(self.model_wrapper)
        self.model_wrapper.trainable_layers = train_layers

        train_params_plugin, self.all_plugin = parse_plugin_cfg(self.model_wrapper, self.cfgs.model_plugin)
        train_params += train_params_plugin

        N_params_unet = format_number(sum(sum(x.numel() for x in p["params"]) for p in train_params))
        self.loggers.info(f"model trainable params: {N_params_unet}")

        return train_params

    def build_optimizer_scheduler(self):
        # set optimizer
        parameters = self.get_param_group_train()

        if len(parameters) > 0:  # do fine-tuning
            cfg_opt = self.cfgs.train.optimizer
            if self.cfgs.train.scale_lr:
                self.scale_lr(parameters)
            assert isinstance(cfg_opt, partial), f"optimizer config should be partial."
            self.optimizer = cfg_opt(params=parameters)
            self.lr_scheduler = get_scheduler(self.cfgs.train.scheduler, self.optimizer, self.cfgs.train.train_steps)

    def train(self, loss_ema=0.93):
        total_batch_size = sum(self.batch_size_list) * self.world_size * self.cfgs.train.gradient_accumulation_steps

        self.loggers.info("***** Running training *****")
        self.loggers.info(
            f"  Num batches each epoch = {[len(loader) for loader in self.train_loader_group.loader_dict.values()]}"
        )
        self.loggers.info(f"  Num Steps = {self.cfgs.train.train_steps}")
        self.loggers.info(f"  Instantaneous batch size per device = {sum(self.batch_size_list)}")
        self.loggers.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_batch_size}")
        self.loggers.info(f"  Gradient Accumulation steps = {self.cfgs.train.gradient_accumulation_steps}")
        self.global_step = 0
        if self.cfgs.train.resume is not None:
            self.global_step = self.cfgs.train.resume.start_step
            self.lr_scheduler.step(self.cfgs.train.resume.start_step)

        self.model_wrapper.train()
        loss_sum = None
        for data_dict in self.train_loader_group:
            loss, pred_dict, inputs_dict = self.train_one_step(data_dict)
            loss_sum = loss if loss_sum is None else (loss_ema * loss_sum + (1 - loss_ema) * loss)

            self.global_step += 1
            if self.is_local_main_process:
                if self.global_step % self.cfgs.train.save_step == 0:
                    self.save_model()
                if self.global_step % self.min_log_step == 0:
                    # get learning rate from optimizer
                    lr_model = self.optimizer.param_groups[0]["lr"] if hasattr(self, "optimizer") else 0.0
                    log_data = {
                        "train/Step": {
                            "format": "[{}/{}]",
                            "data": [self.global_step, self.cfgs.train.train_steps],
                        },
                        "train/Epoch": {
                            "format": "[{}/{}]<{}/{}>",
                            "data": [
                                self.global_step // self.steps_per_epoch,
                                self.cfgs.train.train_epochs,
                                self.global_step % self.steps_per_epoch,
                                self.steps_per_epoch,
                            ],
                        },
                        "train/LR_model": {"format": "{:.2e}", "data": [lr_model]},
                        "train/Loss": {"format": "{:.5f}", "data": [loss_sum]},
                    }
                    if self.metric_train is not None:
                        if is_dict(self.metric_train):
                            for ds_name in pred_dict.keys():
                                if self.metric_train[ds_name] is not None:
                                    self.metric_train[ds_name].reset()
                                    self.metric_train[ds_name].update(pred_dict[ds_name], inputs_dict[ds_name])
                                    metrics_dict = self.metric_train[ds_name].finish(None, self.is_local_main_process)
                                    log_data.update(MetricGroup.format(metrics_dict, prefix=f'{ds_name}/'))
                        else:
                            self.metric_train.reset()
                            for ds_name in pred_dict.keys():
                                self.metric_train.update(pred_dict[ds_name], inputs_dict[ds_name])
                            metrics_dict = self.metric_train.finish(None, self.is_local_main_process)
                            log_data.update(MetricGroup.format(metrics_dict))
                    self.loggers.log(
                        datas=log_data,
                        step=self.global_step,
                    )

            if self.evaluator is not None:
                self.evaluator.evaluate(self.global_step, self.model_wrapper)
                self.model_wrapper.train()

            if self.global_step >= self.cfgs.train.train_steps:
                break

        self.wait_for_everyone()
        if self.is_local_main_process:
            self.save_model()

    def train_one_step(self, data_dict):

        v_proc = lambda v: v.detach() if isinstance(v, torch.Tensor) else v

        pred_dict, inputs_dict = {}, {}
        loss_all = []
        with self.accelerator.accumulate(self.model_wrapper):
            for ds_name, data in data_dict.items():
                input_datas = {k: self.to_dev(v) for k, v in data.items() if k != "plugin_input"}
                if "plugin_input" in data:
                    input_datas["plugin_input"] = {k: self.to_dev(v) for k, v in data["plugin_input"].items()}

                model_pred = self.model_wrapper(ds_name, **input_datas)
                pred_dict[ds_name] = {k: v_proc(v) for k, v in model_pred.items()}
                inputs_dict[ds_name] = {k: v_proc(v) for k, v in input_datas.items()}
                loss = self.get_loss(ds_name, model_pred, input_datas)
                loss_all.append(loss.item())
                self.accelerator.backward(loss, retain_graph=self.cfgs.train.retain_graph)

            if hasattr(self, "optimizer"):
                if self.cfgs.train.max_grad_norm and self.accelerator.sync_gradients:  # fine-tuning
                    if hasattr(self.model_wrapper, "trainable_parameters"):
                        clip_param = self.model_wrapper.trainable_parameters
                    else:
                        clip_param = self.model_wrapper.module.trainable_parameters
                    self.accelerator.clip_grad_norm_(clip_param, self.cfgs.train.max_grad_norm)
                self.optimizer.step()
                if self.lr_scheduler:
                    self.lr_scheduler.step()
                self.optimizer.zero_grad(set_to_none=self.cfgs.train.set_grads_to_none)

            self.model_raw.update_model(self.global_step)  # Some model may update by step
            self.update_ema()
        return sum(loss_all), pred_dict, inputs_dict

    def get_loss(self, ds_name, model_pred, inputs):
        weight = self.train_loader_group.get_loss_weights(ds_name)
        if is_dict(self.criterion):
            loss = self.criterion[ds_name](model_pred, inputs).mean()
        else:
            loss = self.criterion(model_pred, inputs).mean()
        return loss * weight

    def update_ema(self):
        if hasattr(self, "ema_model"):
            self.ema_model.step(self.model_raw.named_parameters())

    def save_model(self, from_raw=False):
        self.ckpt_manager.save(
            name=self.cfgs.model.name,
            step=self.global_step,
            model=self.model_raw,
            all_plugin=self.all_plugin,
            ema=getattr(self, "ema_model", None),
        )

        self.loggers.info(f"Saved state, step: {self.global_step}")

    def wait_for_everyone(self):
        self.accelerator.wait_for_everyone()


def neko_train():
    import subprocess
    import sys
    subprocess.run(["accelerate", "launch", "-m", "rainbowneko.train.trainer.trainer_ac"] + sys.argv[1:])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Stable Diffusion Training")
    parser.add_argument("--cfg", type=str, default=None, required=True)
    args, cfg_args = parser.parse_known_args()

    parser, conf = load_config_with_cli(args.cfg, args_list=cfg_args)  # skip --cfg
    trainer = Trainer(parser, conf)
    trainer.train()



from functools import partial
from typing import Optional, Union

import torch
from torch import nn
from torch.optim import lr_scheduler

from .lr_scheduler import SchedulerType, TYPE_TO_SCHEDULER_FUNCTION, Optimizer


def get_scheduler(cfg, optimizer, num_training_steps):
    if cfg is None:
        return None
    elif isinstance(cfg, partial):
        return cfg(optimizer=optimizer)
    else:
        return get_scheduler_with_name(optimizer=optimizer, num_training_steps=num_training_steps, **cfg)

def get_scheduler_with_name(
    name: Union[str, SchedulerType],
    optimizer: Optimizer,
    num_warmup_steps: Optional[int] = None,
    num_training_steps: Optional[int] = None,
    **kwargs
):
    """
    Unified API to get any scheduler from its name.

    Args:
        name (`str` or `SchedulerType`):
            The name of the scheduler to use.
        optimizer (`torch.optim.Optimizer`):
            The optimizer that will be used during training.
        num_warmup_steps (`int`, *optional*):
            The number of warmup steps to do. This is not required by all schedulers (hence the argument being
            optional), the function will raise an error if it's unset and the scheduler type requires it.
        num_training_steps (`int``, *optional*):
            The number of training steps to do. This is not required by all schedulers (hence the argument being
            optional), the function will raise an error if it's unset and the scheduler type requires it.
        num_cycles (`int`, *optional*):
            The number of hard restarts used in `COSINE_WITH_RESTARTS` scheduler.
        power (`float`, *optional*, defaults to 1.0):
            Power factor. See `POLYNOMIAL` scheduler
        last_epoch (`int`, *optional*, defaults to -1):
            The index of the last epoch when resuming training.
    """
    # All other schedulers require `num_training_steps`
    if num_training_steps is None:
        raise ValueError(f"{name} requires `num_training_steps`, please provide that argument.")

    # All other schedulers require `num_warmup_steps`
    if num_warmup_steps is None:
        raise ValueError(f"{name} requires `num_warmup_steps`, please provide that argument.")

    if isinstance(num_warmup_steps, float): # warmup ratio
        num_warmup_steps = int(num_warmup_steps * num_training_steps)

    # One Cycle for super convergence
    if name == 'one_cycle':
        scheduler = lr_scheduler.OneCycleLR(optimizer, max_lr=[x['lr'] for x in optimizer.state_dict()['param_groups']],
                                            steps_per_epoch=num_training_steps, epochs=1,
                                            pct_start=num_warmup_steps/num_training_steps, **kwargs)
        return scheduler

    name = SchedulerType(name)
    schedule_func = TYPE_TO_SCHEDULER_FUNCTION[name]
    if name == SchedulerType.CONSTANT:
        return schedule_func(optimizer, **kwargs)

    if name == SchedulerType.CONSTANT_WITH_WARMUP:
        return schedule_func(optimizer, num_warmup_steps=num_warmup_steps, **kwargs)

    if name == SchedulerType.COSINE_WITH_RESTARTS:
        return schedule_func(
            optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=num_training_steps, **kwargs
        )

    if name == SchedulerType.POLYNOMIAL:
        return schedule_func(
            optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=num_training_steps, **kwargs
        )

    return schedule_func(optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=num_training_steps, **kwargs)

def remove_all_hooks(model: nn.Module) -> None:
    for name, child in model.named_modules():
        child._forward_hooks.clear()
        child._forward_pre_hooks.clear()
        child._backward_hooks.clear()

def remove_layers(model: nn.Module, layer_class):
    named_modules = {k:v for k, v in model.named_modules()}
    for k, v in named_modules.items():
        if isinstance(v, layer_class):
            parent, name = named_modules[k.rsplit('.', 1)]
            delattr(parent, name)
            del v

def hook_compile(model):
    named_modules = {k:v for k, v in model.named_modules()}

    for name, block in named_modules.items():
        if len(block._forward_hooks)>0:
            for hook in block._forward_hooks.values():  # 从前往后执行
                old_forward = block.forward

                def new_forward(*args, **kwargs):
                    result = old_forward(*args, **kwargs)
                    hook_result = hook(block, args, result)
                    if hook_result is not None:
                        result = hook_result
                    return result

                block.forward = new_forward

        if len(block._forward_pre_hooks)>0:
            for hook in list(block._forward_pre_hooks.values())[::-1]:  # 从前往后执行
                old_forward = block.forward

                def new_forward(*args, **kwargs):
                    result = hook(block, args)
                    if result is not None:
                        if not isinstance(result, tuple):
                            result = (result,)
                    else:
                        result = args
                    return old_forward(*result, **kwargs)

                block.forward = new_forward
    remove_all_hooks(model)

def _convert_cpu(t):
    return t.to('cpu') if t.device.type == 'cuda' else t

def _convert_cuda(t):
    return t.to('cuda') if t.device.type == 'cpu' else t

def to_cpu(model):
    model._apply(_convert_cpu)
    torch.cuda.empty_cache()
    torch.cuda.synchronize()

def to_cuda(model):
    model._apply(_convert_cuda)

def split_module_name(layer_name):
    name_split = layer_name.rsplit('.', 1)
    if len(name_split) == 1:
        parent_name, host_name = '', name_split[0]
    else:
        parent_name, host_name = name_split
    return parent_name, host_name



import os
from typing import Any, Dict

from addict import Dict as ADict
from rainbowneko.utils.img_size_tool import types_support
from rainbowneko.utils.utils import get_file_ext

from .base import VisionDataSource


class ImageFolderClassSource(VisionDataSource):
    def __init__(self, img_root, repeat=1, use_cls_index=True, **kwargs):
        super(ImageFolderClassSource, self).__init__(img_root, repeat=repeat, **kwargs)

        self.use_cls_index = use_cls_index
        self.img_paths, self.label_dict = self._load_img_label(img_root)

    def __len__(self):
        return len(self.img_paths)

    def __getitem__(self, index) -> Dict[str, Any]:
        path = self.img_paths[index]
        return {
            'image': path,
            'label': self.load_label(path)
        }

    def _load_img_label(self, img_root):
        sub_folders = [os.path.join(img_root, x) for x in os.listdir(img_root)]
        class_imgs = []
        label_dict = ADict()
        cls_id_dict = {}
        for class_folder in sub_folders:
            class_name = os.path.basename(class_folder)
            imgs = []
            for x in os.listdir(class_folder):
                if get_file_ext(x) in types_support:
                    imgs.append(os.path.join(class_folder, x))
                    if self.use_cls_index:
                        if class_name not in cls_id_dict:
                            cls_id_dict[class_name] = len(cls_id_dict)
                        label_dict[class_name][x] = cls_id_dict[class_name]
                    else:
                        label_dict[class_name][x] = class_name
            if isinstance(self.repeat, int):
                class_imgs.extend(imgs * self.repeat)
            else:
                class_imgs.extend(imgs * self.repeat[class_name])
        if self.use_cls_index:  # record class name to index map
            self.cls_id_dict = cls_id_dict
        return class_imgs, label_dict

    def load_label(self, path: str) -> Dict[str, Any]:
        img_root, img_name = os.path.split(path)
        img_root, class_name = os.path.split(img_root)

        label = self.label_dict[class_name].get(img_name, None)
        return label



from typing import Iterable, Tuple, Dict

import torch
from torch import nn
from copy import deepcopy


class ModelEMA:
    def __init__(self, model: nn.Module, decay_max=0.997, decay_factor=(1, 5),
                 optimization_step=0, interval=1):
        self.train_params = {name: p.clone().detach() for name, p in model.state_dict().items()}
        self.model = deepcopy(model)
        model.load_state_dict(self.train_params)

        self.decay_max = decay_max
        self.decay_factor = decay_factor
        self.optimization_step = optimization_step
        self.interval = interval

    @torch.no_grad()
    def step(self, model: nn.Module, step: int = 0):
        self.optimization_step += 1

        if step % self.interval == 0:
            # Compute the decay factor for the exponential moving average.
            if self.decay_factor[0] == self.decay_factor[1]:
                one_minus_decay = 1 - self.decay_max
            else:
                value = (self.decay_factor[0] + self.optimization_step) / (self.decay_factor[1] + self.optimization_step)
                one_minus_decay = 1 - min(self.decay_max, value)

            for name, param in model.state_dict().items():
                if name in self.train_params:
                    s_param = self.train_params[name]
                    if torch.is_floating_point(s_param):
                        s_param.sub_(one_minus_decay * (s_param - param.to(s_param.device)))

        # torch.cuda.empty_cache()

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def copy_to(self, parameters: Iterable[Tuple[str, torch.nn.Parameter]]) -> None:
        for name, param in parameters:
            if name in self.train_params:
                param.data.copy_(self.train_params[name])

    def to(self, device=None, dtype=None):
        # .to() on the tensors handles None correctly
        self.train_params = {
            name: (p.to(device=device, dtype=dtype) if p.is_floating_point() else p.to(device=device))
            for name, p in self.train_params
        }
        return self

    def state_dict(self) -> Dict[str, torch.Tensor]:
        return self.train_params

    def load_state_dict(self, state: Dict[str, torch.Tensor], prefix=None):
        for k, v in state:
            if prefix is None:
                if k in self.train_params:
                    self.train_params[k] = v
            else:
                if k in self.train_params and k.startswith(prefix):
                    self.train_params[k] = v



from .base import BasicAction, MemoryMixin, LoopAction, feedback_input
from .io import LoadImageAction, LoadModelAction
from .model import ForwardAction, VisPredAction, BuildModelAction, PrepareAction


import ast
import importlib
import inspect
import os
import shutil

from omegaconf import OmegaConf

from .yaml_cfg import YamlCfgParser


class CallTransformer(ast.NodeTransformer):
    transform_parent = (ast.Call, ast.Expr, ast.Dict, ast.List)

    def __init__(self):
        self.parent_stack = []

    def visit(self, node):
        # 在遍历到每个节点时，将当前节点推入栈
        self.parent_stack.append(node)
        result = super().visit(node)
        # 完成当前节点的处理后，将其从栈中弹出
        self.parent_stack.pop()
        return result

    def visit_Call(self, node):
        # 创建一个新的Call节点，调用dict函数并传递关键字参数
        call_node = ast.Call(
            func=ast.Name(id='dict', ctx=ast.Load()),
            args=[],
            keywords=[]
        )

        partial_flag = False

        # skip node that cannot transform
        parent_node = self.parent_stack[-2] if len(self.parent_stack) > 1 else None
        if not isinstance(parent_node, self.transform_parent):
            return node

        if isinstance(node.func, ast.Attribute):
            call_node.keywords.append(
                ast.keyword(arg='_target_', value=ast.Attribute(attr=node.func.attr, value=node.func.value)))
        else:
            if node.func.id == 'partial':
                partial_flag = True
            elif not node.func.id == 'dict':
                call_node.keywords.append(ast.keyword(arg='_target_', value=ast.Name(id=node.func.id)))


        # 处理位置参数
        if node.args:
            args_list = [self.visit(arg) for arg in node.args]
            if partial_flag:
                call_node.keywords.append(ast.keyword(arg='_target_', value=args_list[0]))
                call_node.keywords.append(ast.keyword(arg='_partial_', value=ast.NameConstant(value=True)))
                if len(args_list) > 1:
                    args_list = ast.List(elts=args_list[1:], ctx=ast.Load())
                    call_node.keywords.append(ast.keyword(arg='_args_', value=args_list))
            else:
                # 创建一个列表节点，包含所有的位置参数
                args_list = ast.List(elts=args_list, ctx=ast.Load())
                # 将列表节点作为字典的值
                call_node.keywords.append(ast.keyword(arg='_args_', value=args_list))

        # 处理关键字参数
        if node.keywords:
            for keyword in node.keywords:
                call_node.keywords.append(ast.keyword(arg=keyword.arg, value=self.visit(keyword.value)))

        # 替换原始的Call节点
        return call_node


class PythonCfgParser(YamlCfgParser):
    def __init__(self):
        super().__init__()
        self.cfg_dict = {}

    def get_code(self, func):
        # 获取函数源代码
        source_code = inspect.getsource(func)

        # 分离函数定义和函数体
        start_index = source_code.find(':\n') + 2  # 找到第一个换行符后的索引
        function_body = source_code[start_index:].strip()
        return function_body

    def transform_code(self, code):
        # 解析代码为AST
        tree = ast.parse(code)
        # 应用转换器
        transformer = CallTransformer()
        new_tree = transformer.visit(tree)

        # 将AST转换回代码字符串
        new_code = ast.unparse(new_tree)

        # self.print_code(new_code)

        return new_code

    def print_code(self, code):
        # 使用yapf格式化代码
        from yapf.yapflib.yapf_api import FormatCode
        new_code, _ = FormatCode(code, style_config='facebook')
        print(new_code)

    def resolve_sub_cfgs(self, module, cfg):
        if isinstance(cfg, dict):
            if '_target_' in cfg and getattr(cfg['_target_'], '_neko_cfg_', False):
                code = self.get_code(cfg['_target_'])
                code_format = self.transform_code(code)
                del cfg['_target_']
                cfg = eval(code_format, vars(module), cfg)
                return cfg

            for key, value in cfg.items():
                if isinstance(value, dict):
                    res = self.resolve_sub_cfgs(module, value)
                    if res is not None:
                        cfg[key] = res
                if isinstance(value, list):
                    self.resolve_sub_cfgs(module, value)
        elif isinstance(cfg, list):
            for idx, value in enumerate(cfg):
                if isinstance(value, dict):
                    res = self.resolve_sub_cfgs(module, value)
                    if res is not None:
                        cfg[idx] = res
                elif isinstance(value, list):
                    self.resolve_sub_cfgs(module, value)

    def load_cfg(self, path: str, trans=True):
        # record for save
        if len(self.cfg_dict) == 0:
            self.cfg_dict['cfg.py'] = path
        else:
            self.cfg_dict[path] = path

        # load module
        module_name = os.path.splitext(os.path.basename(path))[0]
        spec = importlib.util.spec_from_file_location(module_name, path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        if trans:
            code = self.get_code(module.make_cfg)
            code_format = self.transform_code(code)
            cfg = eval(code_format, vars(module))
        else:
            cfg = module.config
        self.resolve_sub_cfgs(module, cfg)

        return OmegaConf.create(cfg, flags={"allow_objects": True})

    def save_configs(self, cfg, path):
        for dst, src in self.cfg_dict.items():
            path_dst = os.path.join(path, dst)
            os.makedirs(os.path.dirname(path_dst), exist_ok=True)
            shutil.copy2(src, path_dst)



"""
ckpt_pkl.py
====================
    :Name:        save model with torch
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     8/04/2023
    :Licence:     MIT
"""

import os
from typing import Dict, Union

import torch
from torch import nn

from rainbowneko.models.plugin import PluginGroup, BasePluginBlock
from rainbowneko.models.wrapper import BaseWrapper
from .base import CkptManagerBase


class CkptManagerPKL(CkptManagerBase):
    def __init__(self, plugin_from_raw=False, saved_model=({'model':'', 'trainable':True},), **kwargs):
        self.plugin_from_raw = plugin_from_raw
        self.saved_model = saved_model

    def set_save_dir(self, save_dir):
        os.makedirs(save_dir, exist_ok=True)
        self.save_dir = save_dir

    def exclude_state(self, state, key):
        if key is None:
            return state
        else:
            return {k: v for k, v in state.items() if key not in k}

    def save_model(self, model: nn.Module, name, step, model_ema=None, exclude_key=None):
        sd_base = {}
        for item in self.saved_model:
            block = model if item['model']=='' else eval(f"model.{item['model']}")
            sd_base.update(self.exclude_state(
                BasePluginBlock.extract_state_without_plugin(block, trainable=item['trainable']), exclude_key
            ))

        sd_model = {"base": sd_base}
        if model_ema is not None:
            sd_ema = model_ema.state_dict()
            sd_ema = {k: sd_ema[k] for k in sd_model["base"].keys()}
            sd_model["base_ema"] = self.exclude_state(sd_ema, exclude_key)
        self._save_ckpt(sd_model, name, step)

    def save_plugins(
            self, host_model: nn.Module, plugins: Dict[str, PluginGroup], name: str, step: int, model_ema=None
    ):
        if len(plugins) > 0:
            sd_plugin = {}
            for plugin_name, plugin in plugins.items():
                sd_plugin["plugin"] = plugin.state_dict(host_model if self.plugin_from_raw else None)
                if model_ema is not None:
                    sd_plugin["plugin_ema"] = plugin.state_dict(model_ema)
                self._save_ckpt(sd_plugin, f"{name}-{plugin_name}", step)

    def _save_ckpt(self, sd_model, name=None, step=None, save_path=None):
        if save_path is None:
            save_path = os.path.join(self.save_dir, f"{name}-{step}.ckpt")
        torch.save(sd_model, save_path)

    def load_ckpt(self, ckpt_path, map_location="cpu"):
        return torch.load(ckpt_path, map_location=map_location)

    def load_ckpt_to_model(self, model: nn.Module, ckpt_path, model_ema=None):
        sd = self.load_ckpt(ckpt_path)
        if "base" in sd:
            for item in self.saved_model:
                block = model if item['model'] == '' else eval(f"model.{item['model']}")
                missing_keys, unexpected_keys = block.load_state_dict(sd["base"], strict=False)
                if len(missing_keys)>0:
                    print(f'missing_keys of model.{item["model"]}: {missing_keys}')
        if "plugin" in sd:
            model.load_state_dict(sd["plugin"], strict=False)

        if model_ema is not None:
            if "base" in sd:
                for item in self.saved_model:
                    prefix = None if item['model'] == '' else item['model']
                    model_ema.load_state_dict(sd["base_ema"], prefix=prefix)
            if "plugin" in sd:
                model_ema.load_state_dict(sd["plugin_ema"])

    def save(self, name, step, model, all_plugin, ema={}, **kwargs):
        """

        :param step: current model train step
        :param model: Model Wrapper for training
        :param all_plugin: all plugins
        :param ema: model or plugins ema {'model':..., 'plugin':...}
        :return:
        """
        if len(kwargs) > 0:
            print(f'Unused kwargs in save model: {", ".join(kwargs.keys())}')

        self.save_model(model, model_ema=getattr(ema, "model", None), name=name, step=step)
        self.save_plugins(model, all_plugin, name=name, step=step, model_ema=getattr(self, "model", None))

    @classmethod
    def load(cls, pretrained_model, **kwargs):
        if len(kwargs) > 0:
            print(f'Unused kwargs in load model: {", ".join(kwargs.keys())}')

        raise NotImplementedError(f"{cls} dose not support load()")

    @classmethod
    def load_to_model(cls, model: BaseWrapper, ckpt_path: Union[str, Dict[str, str]], **kwargs):
        if len(kwargs) > 0:
            print(f'Unused kwargs in load model: {", ".join(kwargs.keys())}')

        manager = cls()

        # rainbowneko format model ckpt is in 'base' key.
        def load_ckpt(ckpt_path):
            ckpt = manager.load_ckpt(ckpt_path)
            if 'base' in ckpt:
                return ckpt['base']
            else:
                return ckpt

        if isinstance(ckpt_path, str):
            model.load_state_dict(load_ckpt(ckpt_path))
        else:
            for k, v in ckpt_path.items():
                eval(f'model.{k}').load_state_dict(load_ckpt(v))

        return model



from typing import List, Dict
from tqdm.auto import tqdm

def feedback_input(fun):
    def f(*args, **states):
        output = fun(*args, **states)
        if 'memory' in states:
            del states['memory']
        if output is not None:
            states.update(output)
        return states
    return f

class BasicAction:
    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(self, *args, **kwargs):
        raise NotImplementedError()

class MemoryMixin:
    pass

class LoopAction(BasicAction, MemoryMixin):
    def __init__(self, loop_value:Dict[str, str], actions:List[BasicAction]):
        self.loop_value = loop_value
        self.actions = actions

    def forward(self, memory, **states):
        loop_data = [states.pop(k) for k in self.loop_value.keys()]
        pbar = tqdm(zip(*loop_data), total=len(loop_data[0]))
        N_steps = len(self.actions)
        for data in pbar:
            feed_data = {k:v for k,v in zip(self.loop_value.values(), data)}
            states.update(feed_data)
            for step, act in enumerate(self.actions):
                pbar.set_description(f'[{step+1}/{N_steps}] action: {type(act).__name__}')
                if isinstance(act, MemoryMixin):
                    states = act(memory=memory, **states)
                else:
                    states = act(**states)
        return states



import os.path

import torch
from tqdm.auto import tqdm

from rainbowneko.models.wrapper import BaseWrapper
from rainbowneko.train.data import DataGroup
from .evaluator import Evaluator
from .renderer import Renderer


class Previewer(Evaluator):
    def __init__(self, trainer: "Trainer", data_loader_group: DataGroup, img_dir: str, renderer: Renderer, interval=100):
        self.data_loader_group = data_loader_group
        self.trainer = trainer
        self.img_dir = img_dir
        self.renderer = renderer
        self.interval = interval

    def forward_one_step(self, model, data):
        device = self.trainer.device
        weight_dtype = self.trainer.weight_dtype

        image = data.pop("image").to(device, dtype=weight_dtype)
        target = {k: v.to(device) for k, v in data.pop("label").items()}
        other_datas = {
            k: v.to(device, dtype=weight_dtype) for k, v in data.items() if k != "plugin_input"
        }
        if "plugin_input" in data:
            other_datas["plugin_input"] = {
                k: v.to(device, dtype=weight_dtype) for k, v in data["plugin_input"].items()
            }

        model_pred = model(image, **other_datas)

        return model_pred, target

    def save_images(self, pred, target, img_dir, step):
        img_list = self.renderer(pred, target)
        for img in img_list:
            img.save(os.path.join(img_dir, f"{step}-{self.img_count}.png"))
            self.img_count += 1

    @torch.inference_mode()
    def evaluate(self, step: int, model: BaseWrapper):
        if step % self.interval != 0:
            return

        model.eval()

        img_dir = os.path.join(self.trainer.exp_dir, self.img_dir)
        self.trainer.loggers.info(f'Preview to {img_dir}')
        self.img_count=0
        for loader in self.data_loader_group.loader_dict.values():
            for data in tqdm(loader, disable=not self.trainer.is_local_main_process):
                pred, target = self.forward_one_step(model, data)
                self.save_images(pred, target, img_dir=img_dir, step=step)

    def to(self, device):
        self.renderer.to(device)



from typing import Dict, Any, List, Union
import torch

from PIL import Image
from .base import BasicAction, feedback_input, MemoryMixin
from rainbowneko.ckpt_manager import auto_manager

class LoadImageAction(BasicAction):
    def __init__(self, image_paths:Union[str, List[str]], image_transforms=None):
        super().__init__()
        if isinstance(image_paths, str):
            image_paths = [image_paths]
        self.image_paths = image_paths
        self.image_transforms = image_transforms

    def load_one(self, path:str):
        img = Image.open(path).convert('RGB')
        if self.image_transforms:
            img = self.image_transforms(img)
        return img

    @feedback_input
    def forward(self, device, **states):
        input = torch.stack([self.load_one(path) for path in self.image_paths]).to(device)
        input: Dict[str, Any] = {'x':input}
        return {'input':input}

class LoadModelAction(BasicAction, MemoryMixin):
    def __init__(self, part_paths: Union[str, Dict[str, str]]):
        super().__init__()
        self.part_paths=part_paths

    @feedback_input
    def forward(self, memory, **states):
        if isinstance(self.part_paths, str):
            manager = auto_manager(self.part_paths)
        else:
            manager = auto_manager(next(iter(self.part_paths.values())))
        manager.load_to_model(memory.model, self.part_paths)